import 'dart:typed_data';

import 'package:binarize/binarize.dart';

class _Nullable<T> extends PayloadType<T?> {
  const _Nullable(this.type);

  final PayloadType<T> type;

  @override
  int length(T? value) => 1 + (value != null ? type.length(value) : 0);

  @override
  T? get(ByteData data, int offset) {
    final isNull = data.getUint8(offset) == 0;
    return isNull ? null : type.get(data, offset + 1);
  }

  @override
  void set(T? value, ByteData data, int offset) {
    data.setUint8(offset, value == null ? 0 : 1);
    if (value != null) {
      type.set(value, data, offset + 1);
    }
  }
}

/// Make the given [type] nullable.
///
/// When the value is `null`, it will store a single byte with the value `0`.
/// If the value is not `null` it will store a single byte with the value `1`
/// and the encoded result of the [type] right after it.
///
/// It has a byte length of `1 + type.length(value)`.
///
/// Usage:
///
/// ```dart
/// payload.set(nullable(uint8), null);
/// ```
PayloadType<T?> nullable<T>(PayloadType<T> type) => _Nullable(type);
